import functools
import glob
import inspect
import logging
import os
import shutil
import subprocess
import tempfile
import warnings

from contextlib import closing
from multiprocessing import Pool
import pandas as pd
import youtube_dl

from dcase_util.containers import AudioContainer
from desed.utils import create_folder, download_file_from_url
from tqdm import tqdm
from youtube_dl import DownloadError
from youtube_dl.utils import ExtractorError

from .logger import create_logger, DesedWarning
from .utils import create_folder, download_file_from_url


class LoggerYtdlWarnings(object):
    """ Class needed to avoid printing warnings from youtube_dl which cause breaks in the progress bar."""

    def debug(self, msg):
        pass

    def warning(self, msg):
        pass

    def error(self, msg):
        pass


def _download_audioset_file(filename, result_dir, platform="youtube", tmp_folder="tmp"):
    """ download a file from youtube given an audioSet filename. (It takes only a part of the file thanks to
    information provided in the filename)
    Args:
        filename : str, AudioSet filename to download.
        result_dir : str, result directory which will contain the downloaded file.
        platform: str, name of the platform, here youtube or vimeo.
        tmp_folder: str, the directory in which to download the temporary file generated by youtube_dl.
    Return:
        list, Empty list if the file is downloaded, otherwise contains the filename and the error associated

    """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    tmp_filename = ""
    fname_no_ext = os.path.splitext(filename)[0]
    segment_start = fname_no_ext.split("_")[-2]
    segment_end = fname_no_ext.split("_")[-1]
    audio_container = AudioContainer()

    # Define download parameters
    ydl_opts = {
        "format": "bestaudio/best",
        "outtmpl": tmp_folder + "%(id)s.%(ext)s",
        "noplaylist": True,
        "quiet": True,
        "prefer_ffmpeg": True,
        "logger": LoggerYtdlWarnings(),
        "audioformat": "wav",
    }

    if platform.lower() == "youtube":
        query_id = filename[1:12]  # Remove the Y in front of the file.
        baseurl = "https://www.youtube.com/watch?v="
    elif platform.lower() == "vimeo":
        query_id = filename.split("_")[0]
        baseurl = "https://vimeo.com/"
    else:
        raise NotImplementedError("platform can only be vimeo or youtube")

    if not os.path.isfile(os.path.join(result_dir, filename)):
        try:
            logger.debug(filename)
            # Download file
            with youtube_dl.YoutubeDL(ydl_opts) as ydl:
                meta = ydl.extract_info(f"{baseurl}{query_id}", download=True)

            audio_formats = [f for f in meta["formats"] if f.get("vcodec") == "none"]
            if audio_formats is []:
                return [filename, "no audio format available"]
            # get the best audio format
            best_audio_format = audio_formats[-1]

            tmp_filename = tmp_folder + query_id + "." + best_audio_format["ext"]
            audio_container.load(
                filename=tmp_filename,
                fs=44100,
                res_type="kaiser_best",
                start=float(segment_start),
                stop=float(segment_end),
                auto_trimming=True,
            )

            # Save segmented audio
            audio_container.filename = filename
            audio_container.detect_file_format()
            audio_container.save(filename=os.path.join(result_dir, filename))

            # Remove temporary file
            os.remove(tmp_filename)
            return []

        except (KeyboardInterrupt, SystemExit):
            # Remove temporary files and current audio file.
            for fpath in glob.glob(tmp_folder + query_id + "*"):
                os.remove(fpath)
            raise

        # youtube-dl error, file often removed, IO Error is for AudioContainer error if length of file is different.
        except (ExtractorError, DownloadError, IOError) as e:
            if os.path.exists(tmp_filename):
                os.remove(tmp_filename)

            return [filename, str(e)]

        # multiprocessing can give this error
        except (IndexError, ValueError) as e:
            if os.path.exists(tmp_filename):
                os.remove(tmp_filename)
            logger.info(filename)
            logger.info(str(e))
            return [filename, "Index Error"]
    else:
        logger.debug(filename, "exists, skipping")
        return []


def download_audioset_files(
    filenames,
    result_dir,
    n_jobs=1,
    chunk_size=10,
    missing_files_tsv="..",
    platform="youtube",
):
    """ download files in parallel from youtube given a tsv file listing files to download.
    It also stores not downloaded files with their associated error in "missing_files_[tsv_file].tsv"

       Args:
           filenames : pandas Series, named "filename" listing AudioSet filenames to download
           result_dir : str, result directory which will contain downloaded files
           n_jobs : int, number of download to execute in parallel
           chunk_size : int, number of files to download before updating the progress bar. Bigger it is, faster it goes
                because data is filled in memory but progress bar only updates after a chunk is finished.
           missing_files_tsv: str, path of the tsv which will contain the files that couldn't have been downloaded.
           platform: str, the platform the filenames are coming from "youtube" or "vimeo"

       Returns:
           missing_files : pandas.DataFrame, files not downloaded whith associated error.

       """
    warnings.filterwarnings("ignore")
    create_folder(result_dir)
    TMP_FOLDER = "tmp/"
    create_folder(TMP_FOLDER)

    p = None
    files_error = []
    try:
        if n_jobs == 1:
            for filename in tqdm(filenames):
                files_error.append(
                    _download_audioset_file(filename, result_dir, platform)
                )
        # multiprocessing
        else:
            with closing(Pool(n_jobs)) as p:
                # Put result_dir and platform as constants variable with result_dir in download_file
                download_file_alias = functools.partial(
                    _download_audioset_file,
                    result_dir=result_dir,
                    platform=platform,
                    tmp_folder=TMP_FOLDER,
                )

                for val in tqdm(
                    p.imap_unordered(download_file_alias, filenames, chunk_size),
                    total=len(filenames),
                ):
                    files_error.append(val)

        # Store files which gave error
        missing_files = pd.DataFrame(files_error).dropna()
        if not missing_files.empty:
            # Save missing_files to be able to ask them
            missing_files.columns = ["filename", "error"]
            missing_files.to_csv(missing_files_tsv, index=False, sep="\t")
            warnings.warn(
                f"There are missing files at {missing_files_tsv}, \n"
                f"see info on https://github.com/turpaultn/desed on how to get them",
                DesedWarning,
            )

    except KeyboardInterrupt:
        if p is not None:
            p.terminate()
        raise KeyboardInterrupt

    if os.path.exists(TMP_FOLDER):
        shutil.rmtree(TMP_FOLDER)
    warnings.resetwarnings()
    return missing_files


def download_audioset_files_from_csv(
    tsv_path, result_dir, missing_files_tsv=None, n_jobs=3, chunk_size=10
):
    """ Download audioset files from a tsv_path containing a column "filename"

    Args:
        tsv_path: str, the path to the tsv (tab separated values) containing the column 'filename' with the files to
            download
        result_dir: str, the directory in which to download the files.
        missing_files_tsv: str, path of the tsv which will contain the files that couldn't have been downloaded.
        n_jobs : int, number of download to execute in parallel
        chunk_size : int, number of files to download before updating the progress bar. Bigger it is, faster it goes
            because data is filled in memory but progress bar only updates after a chunk is finished.

    Returns:

    """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    logger.info(f"downloading data from: {tsv_path}")
    if missing_files_tsv is None:
        missing_files_tsv = (
            "missing_files" + os.path.basename(os.path.splitext(tsv_path)[0]) + ".tsv"
        )
    # read metadata file and get only one filename once
    df = pd.read_csv(tsv_path, header=0, sep="\t")
    filenames_test = df["filename"].drop_duplicates()

    download_audioset_files(
        filenames_test,
        result_dir,
        n_jobs=n_jobs,
        chunk_size=chunk_size,
        missing_files_tsv=missing_files_tsv,
    )
    logger.info("###### DONE #######")


def download_eval_public(dataset_folder):
    """ Download the public eval part of desed dataset from Zenodo.

    Args:
        dataset_folder: str, the path to the root of the dataset where to download the evaluation files (this folder
            contains audio and metadata folders).

    Returns:

    """
    create_folder(dataset_folder)
    url_public_eval = (
        f"https://zenodo.org/record/4560759/files/DESED_public_eval.tar.gz?download=1"
    )
    download_and_unpack_archive(url_public_eval, dataset_folder)


def download_audioset_data(
    dataset_folder,
    weak=True,
    unlabel_in_domain=True,
    validation=True,
    n_jobs=3,
    chunk_size=10,
):
    """ Download the DESED dataset files from Audioset.

    Args:
        dataset_folder: str, the path to the root of the dataset where to download the evaluation files (this folder
            contains audio and metadata folders).
        weak: bool, whether to download the weak set or not.
        unlabel_in_domain: bool, whether to download the unlabel_in_domain set or not.
        validation: bool, whether to download the validation set or not.
        n_jobs : int, number of download to execute in parallel
        chunk_size : int, number of files to download before updating the progress bar. Bigger it is, faster it goes
            because data is filled in memory but progress bar only updates after a chunk is finished.

    Returns:

    """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    basedir_missing_files = "missing_files"
    create_folder(basedir_missing_files)
    create_folder(dataset_folder)

    # Metadata:
    url_metadata = (
        f"https://zenodo.org/record/4560857/files/audioset_metadata.tar.gz?download=1"
    )
    download_and_unpack_archive(url_metadata, dataset_folder)

    if weak:
        logger.info("Downloading Weakly labeled data...")
        download_audioset_files_from_csv(
            os.path.join(dataset_folder, "metadata", "train", "weak.tsv"),
            os.path.join(dataset_folder, "audio", "train", "weak"),
            missing_files_tsv=os.path.join(
                basedir_missing_files, "missing_files_" + "weak" + ".tsv"
            ),
            n_jobs=n_jobs,
            chunk_size=chunk_size,
        )

    if unlabel_in_domain:
        logger.info("Downloading Unlabeled (in_domain) labeled data...")
        download_audioset_files_from_csv(
            os.path.join(dataset_folder, "metadata", "train", "unlabel_in_domain.tsv"),
            os.path.join(dataset_folder, "audio", "train", "unlabel_in_domain"),
            missing_files_tsv=os.path.join(
                basedir_missing_files, "missing_files_" + "unlabel_in_domain" + ".tsv"
            ),
            n_jobs=n_jobs,
            chunk_size=chunk_size,
        )

    if validation:
        logger.info("Downloading validation, strongly labeled data...")
        download_audioset_files_from_csv(
            os.path.join(dataset_folder, "metadata", "validation", "validation.tsv"),
            os.path.join(dataset_folder, "audio", "validation"),
            missing_files_tsv=os.path.join(
                basedir_missing_files, "missing_files_" + "validation" + ".tsv"
            ),
            n_jobs=n_jobs,
            chunk_size=chunk_size,
        )

    logger.info(
        f"Please check your missing_files: {basedir_missing_files}, "
        f"you can relaunch 'download_audioset_sets' to try to recude them, "
        "then, send these missing_files to "
    )


def download_real(
    dataset_folder,
    weak=True,
    unlabel_in_domain=True,
    validation=True,
    eval=True,
    n_jobs=3,
    chunk_size=10,
):
    """ Download the DESED real part of the dataset.

    Args:
        dataset_folder: str, the path to the root of the dataset where to download the evaluation files (this folder
            contains audio and metadata folders).
        weak: bool, whether to download the weak set or not.
        unlabel_in_domain: bool, whether to download the unlabel_in_domain set or not.
        validation: bool, whether to download the validation set or not.
        eval: bool, whether to download the public eval set or not.
        n_jobs : int, number of download to execute in parallel
        chunk_size : int, number of files to download before updating the progress bar. Bigger it is, faster it goes
            because data is filled in memory but progress bar only updates after a chunk is finished.

    Returns:
    """
    if eval:
        download_eval_public(dataset_folder)
    download_audioset_data(
        dataset_folder,
        weak=weak,
        unlabel_in_domain=unlabel_in_domain,
        validation=validation,
        n_jobs=n_jobs,
        chunk_size=chunk_size,
    )


def _copy_files_kept(meta_df, input_dir, output_dir):
    """ Fonction to copy some files from input_dir to output_dir.
    The files to copy are stored in a pandas dataframe with a column "filename" and must match a name in the input_dir.

    Args:
        meta_df: pd.Dataframe, the dataframe with a column "filename" of files to keep (filtered filenames).
        input_dir: str, folder containing filenames to match.
        output_dir: str, the destination folder of the files copied

    Returns:

    """
    list_files_available = glob.glob(os.path.join(input_dir, "*"))
    df_files = pd.DataFrame(list_files_available, columns=["filename"])
    df_files["filename"] = df_files.filename.apply(lambda x: os.path.basename(x))

    merged = meta_df.merge(df_files, on="filename")
    merged.apply(
        lambda x: shutil.copy(
            os.path.join(input_dir, x["filename"]),
            os.path.join(output_dir, x["filename"]),
        ),
        axis=1,
    )


def download_and_unpack_archive(url, destination_folder, archive_format="gztar"):
    """ Download and unpack an archive from the internet. Useful for Zenodo archives.

    Args:
        url: str, URL to be download.
        destination_folder: str, the folder in which to extract the content of the archive.
        archive_format: str, the format of the archive to unpack.

    Returns:

    """
    create_folder(destination_folder)
    # not using tempdir because too big files for some /tmp folders
    archive_folder = tempfile.mkdtemp(prefix="tmp_", dir="./")
    path_dl_tar = tempfile.NamedTemporaryFile(
        dir=archive_folder, suffix="." + os.path.splitext(url.split("?")[0])[1]
    ).name
    download_file_from_url(url, path_dl_tar)
    shutil.unpack_archive(path_dl_tar, destination_folder, format=archive_format)
    shutil.rmtree(archive_folder)


def download_sins(destination_folder):
    """ Download SINS database: see https://zenodo.org/record/1247102

    Args:
        destination_folder: str, the folder in which to download tut (will create `DCASE2018-task5-dev` folder in it)

    Returns:
        str, path of TUT extracted database
    """
    logger = create_logger(
        __name__ + "/" + inspect.currentframe().f_code.co_name,
        terminal_level=logging.INFO,
    )
    create_folder(destination_folder)
    zip_file_url_meta = f"https://zenodo.org/record/1247102/files/DCASE2018-task5-dev.meta.zip?download=1"
    download_and_unpack_archive(
        zip_file_url_meta, destination_folder, archive_format="zip"
    )

    for i in range(1, 24):
        logger.info(f"SINS downloading zip {i} / 23 ...")
        zip_file_url = f"https://zenodo.org/record/1247102/files/DCASE2018-task5-dev.audio.{i}.zip?download=1"
        download_and_unpack_archive(
            zip_file_url, destination_folder, archive_format="zip"
        )

    return os.path.join(destination_folder, "DCASE2018-task5-dev")


def filter_sins(
    sins_basedir="DCASE2018-task5-dev",
    destination_folder=os.path.join("background", "sins"),
    classes_kept=["other"],
    rm_original_sins=True,
):
    """ Fonction to copy a only some classes of sins into a new folder.
    Args:
        sins_basedir: str, the path where SINS database has been extracted
        destination_folder: str, the path where to store the filtered SINS version
            (in desed: basedir_soundbank/audio/train/background/sins)
        classes_kept: list, list of classes to keep in the origingal SINS database to use as backgrounds
        rm_original_sins: bool, whether to keep the original database too or not.

    Returns:
    """
    logger = create_logger(
        __name__ + "/" + inspect.currentframe().f_code.co_name,
        terminal_level=logging.INFO,
    )
    logger.info("Filtering SINS...")

    df = pd.read_csv(os.path.join(sins_basedir, "meta.txt"), sep="\t", header=None)
    df["filename"] = df[0].apply(lambda x: os.path.basename(x))
    df = df[df[1].isin(classes_kept)]

    _copy_files_kept(
        df, input_dir=os.path.join(sins_basedir, "audio"), output_dir=destination_folder
    )

    if rm_original_sins:
        shutil.rmtree(sins_basedir)


def download_tut(destination_folder):
    """ Download 'TUT Acoustic scenes 2017, Development dataset', see: https://zenodo.org/record/400515
    Args:
        destination_folder: str, the folder in which to download tut (will create `TUT-acoustic-scenes-2017-development`
            folder in it)

    Returns:
        str, path of extracted TUT database
    """
    logger = create_logger(
        __name__ + "/" + inspect.currentframe().f_code.co_name,
        terminal_level=logging.INFO,
    )
    create_folder(destination_folder)
    zip_meta_tut = f"https://zenodo.org/record/400515/files/TUT-acoustic-scenes-2017-development.meta.zip?download=1"
    download_and_unpack_archive(zip_meta_tut, destination_folder, archive_format="zip")

    for i in range(1, 11):
        logger.info(f"TUT (scenes-2017-dev) downloading zip {i} / 10 ...")
        zip_file_url = (
            f"https://zenodo.org/record/400515/files/"
            f"TUT-acoustic-scenes-2017-development.audio.{i}.zip?download=1"
        )
        download_and_unpack_archive(
            zip_file_url, destination_folder, archive_format="zip"
        )

    return os.path.join(destination_folder, "TUT-acoustic-scenes-2017-development")


def filter_tut(
    tut_basedir="TUT-acoustic-scenes-2017-development",
    destination_folder=os.path.join("background", "tut-scenes-2017-dev"),
    classes_kept=["home", "office", "library"],
    rm_original_tut=True,
):
    """ Fonction to copy a only some classes of TUT into a new folder.
    Args:
        tut_basedir: str, the path where TUT database has been extracted
        destination_folder: str, the path where to store the filtered TUT version
            (in desed: basedir_soundbank/audio/train/background/tut-scenes-2017-dev)
        classes_kept: list, list of classes to be kept from TUT.
        rm_original_tut: bool, whether to keep the original database too or not.

    Returns:
    """
    logger = create_logger(
        __name__ + "/" + inspect.currentframe().f_code.co_name,
        terminal_level=logging.INFO,
    )
    logger.info("Filtering SINS...")

    df = pd.read_csv(os.path.join(tut_basedir, "meta.txt"), sep="\t", header=None)
    df["filename"] = df[0].apply(lambda x: os.path.basename(x))
    df = df[df[1].isin(classes_kept)]

    _copy_files_kept(df, os.path.join(tut_basedir, "audio"), destination_folder)

    if rm_original_tut:
        shutil.rmtree(tut_basedir)


def get_backgrounds_train(basedir_soundbank, sins=True, tut=False, keep_original=False):
    """ Download the background files of the DESED soundbank.
    (These files were not included in the Zenodo because of redistributing license issues.)
    Args:
        basedir_soundbank: str, the base folder of DESED dataset.
        sins: bool, download part of sins database as backgrounds.
        tut: bool, download part of TUT database as backgrounds.
        keep_original: bool, whether to keep the original databases too (sins, tut)

    Returns:
    """
    destination_folder = os.path.join(basedir_soundbank, "audio", "train", "soundbank")
    if sins:
        sins_path = download_sins(destination_folder)
        filter_sins(
            sins_path,
            destination_folder=os.path.join(destination_folder, "background", "sins"),
            rm_original_sins=not keep_original,
        )

    if tut:
        tut_path = download_tut(destination_folder)
        filter_tut(
            tut_path,
            destination_folder=os.path.join(
                destination_folder, "background", "tut-scenes-2017-dev"
            ),
            rm_original_tut=not keep_original,
        )


def download_zenodo_soundbank(destination_folder):
    """ Be careful, there are only the foregrounds of training.
    Args:
        destination_folder: str, the path of the root of the soundbank (will create the structure inside)

    Returns:
    """
    create_folder(destination_folder)
    zip_meta_tut = "https://zenodo.org/record/4307908/files/DESED_synth_soundbank.tar.gz?download=1"
    download_and_unpack_archive(zip_meta_tut, destination_folder)


def split_desed_soundbank_train_val(basedir):
    """ Split the training into training and validation (pre-made, 90%/10%) of backgrounds and foregrounds
    Args:
        basedir: str, path where the soundbank is downloaded (parent folder of "audio")

    Returns:

    """
    logger = create_logger(
        __name__ + "/" + inspect.currentframe().f_code.co_name,
        terminal_level=logging.INFO,
    )
    fname_valid = (
        "https://zenodo.org/record/4307908/files/soundbank_validation.tsv?download=1"
    )
    fpath = os.path.join(basedir, "soundbank_validation.tsv")
    download_file_from_url(fname_valid, fpath)
    df = pd.read_csv(fpath, sep="\t")
    for fpath in df.filepath:
        source_path = os.path.join(basedir, fpath.replace("validation", "train"))
        if os.path.exists(source_path):
            destination_path = os.path.join(basedir, fpath)
            create_folder(os.path.dirname(destination_path))
            shutil.move(source_path, destination_path)
    logger.info("Soundbank splitted in train and validation (90%/10%)")


def unsplit_desed_soundbank(basedir):
    """ UnSplit the the soundbank from training and validation (pre-made, 90%/10%) to training only folder.
    Args:
        basedir: str, path where the soundbank is downloaded (parent folder of "audio")

    Returns:

    """
    logger = create_logger(
        __name__ + "/" + inspect.currentframe().f_code.co_name,
        terminal_level=logging.INFO,
    )
    validation_path = os.path.join(basedir, "audio", "validation")
    for rootdir, subdirs, files in os.walk(validation_path):
        for fname in files:
            source_file = os.path.join(rootdir, fname)
            destination_file = source_file.replace("validation", "train")
            create_folder(os.path.dirname(destination_file))
            shutil.move(source_file, destination_file)
    shutil.rmtree(validation_path)
    logger.info("Unsplitted soundbank, validation moved back to train")


def download_fsd50k(destination_folder, gtruth_only=False):
    """ Download FSD50k dataset from Zenodo.
    Args:
        destination_folder: str, path where the FSD50k is extracted.
    """
    logger = create_logger(
        __name__ + "/" + inspect.currentframe().f_code.co_name,
        terminal_level=logging.INFO,
    )
    create_folder(destination_folder)
    if not gtruth_only:
        archive_folder = os.path.join("tmp_fsd50k")
        create_folder(archive_folder)
        # Train
        for id in ["01", "02", "03", "04", "05", "ip"]:
            logger.info(f"Downloading zip file: FSD50K.dev_audio.z{id}")
            url_dev = f"https://zenodo.org/record/4060432/files/FSD50K.dev_audio.z{id}?download=1"
            download_file_from_url(
                url_dev, os.path.join(archive_folder, f"FSD50K.dev_audio.z{id}")
            )
        logger.info("Unpacking files")
        subprocess.call(
            [
                "zip",
                "-s",
                "0",
                os.path.join(archive_folder, "FSD50K.dev_audio.zip"),
                "--out",
                os.path.join(archive_folder, "unsplit_dev.zip"),
            ]
        )
        shutil.unpack_archive(
            os.path.join(archive_folder, "unsplit_dev.zip"), destination_folder
        )

        # Eval
        for id in ["01", "ip"]:
            logger.info(f"Downloading zip file: FSD50K.eval_audio.z{id}")
            url_eval = f"https://zenodo.org/record/4060432/files/FSD50K.eval_audio.z{id}?download=1"
            download_file_from_url(
                url_eval, os.path.join(archive_folder, f"FSD50K.eval_audio.z{id}")
            )
        logger.info("Unpacking files")
        subprocess.call(
            [
                "zip",
                "-s",
                "0",
                os.path.join(archive_folder, "FSD50K.eval_audio.zip"),
                "--out",
                os.path.join(archive_folder, "unsplit_eval.zip"),
            ]
        )
        shutil.unpack_archive(
            os.path.join(archive_folder, "unsplit_eval.zip"), destination_folder
        )
        shutil.rmtree(archive_folder)

    url_doc = "https://zenodo.org/record/4060432/files/FSD50K.doc.zip?download=1"
    url_gtruth = (
        "https://zenodo.org/record/4060432/files/FSD50K.ground_truth.zip?download=1"
    )
    url_meta = "https://zenodo.org/record/4060432/files/FSD50K.metadata.zip?download=1"
    for url in [url_doc, url_gtruth, url_meta]:
        download_and_unpack_archive(url, destination_folder, archive_format="zip")


def download_fuss(destination_folder):
    """ Download FUSS freesound data. These data are a subset of FSD50k, where each file should be a single event.
    Args:
        destination_folder: str, the folder in which to extract FUSS data.
    """
    url = "https://zenodo.org/record/3743844/files/FUSS_fsd_data.tar.gz?download=1"
    download_and_unpack_archive(url, destination_folder)
    url_doc = (
        "https://zenodo.org/record/4012661/files/FUSS_license_doc.tar.gz?download=1"
    )
    download_and_unpack_archive(url_doc, destination_folder)


def download_desed_soundbank(
    basedir,
    sins_bg=True,
    tut_bg=True,
    split_train_valid=True,
    keep_original_sins_tut=False,
):
    """ Download the soundbank.
    Args:
        basedir: str, the dirname where to store the soundbank.
        sins_bg: bool, whether to download SINS backgrounds (default in 2019-2020).
        tut_bg: bool, whether to download 'TUT Acoustic scenes 2017, Development dataset' backgrounds or not
            (appeared in DCASE task 4 2020).
        split_train_valid: bool, whether to split the training foregrounds and backgrounds in training and validation
            (90%/10%). Default since DCASE task 4 2020.
        keep_original_sins_tut: bool, whether to keep the original versions of TUT and SINS.
    """
    print("downloading soundbank (foregrounds)...")
    download_zenodo_soundbank(basedir)
    print("Downloading backgrounds...")
    get_backgrounds_train(
        basedir, sins_bg, tut_bg, keep_original=keep_original_sins_tut
    )
    if split_train_valid:
        print("Splitting soundbank train into train and validation (90%/10%)...")
        split_desed_soundbank_train_val(basedir)



import logging
import sys
import logging.config


def create_logger(logger_name, terminal_level=logging.INFO):
    """ Create a logger.
    Args:
        logger_name: str, name of the logger
        terminal_level: int, logging level in the terminal
    """
    logging.config.dictConfig(
        {"version": 1, "disable_existing_loggers": False,}
    )
    logger = logging.getLogger(logger_name)
    tool_formatter = logging.Formatter("%(levelname)s - %(name)s - %(message)s")

    if type(terminal_level) is str:
        if terminal_level.lower() == "debug":
            res_terminal_level = logging.DEBUG
        elif terminal_level.lower() == "info":
            res_terminal_level = logging.INFO
        elif "warn" in terminal_level.lower():
            res_terminal_level = logging.WARNING
        elif terminal_level.lower() == "error":
            res_terminal_level = logging.ERROR
        elif terminal_level.lower() == "critical":
            res_terminal_level = logging.CRITICAL
        else:
            res_terminal_level = logging.NOTSET
    else:
        res_terminal_level = terminal_level
    logger.setLevel(res_terminal_level)

    # To avoid duplicates handlers (should have only one)
    if not len(logger.handlers[:]):
        terminal_h = logging.StreamHandler(sys.stdout)
        terminal_h.setLevel(res_terminal_level)
        terminal_h.set_name("stdout")
        terminal_h.setFormatter(tool_formatter)
        logger.addHandler(terminal_h)
    return logger


class DesedWarning(Warning):
    """The root DESED warning class"""

    pass


class DesedError(Exception):
    """The root DESED exception class"""

    pass



# -*- coding: utf-8 -*-
"""class and functions to generate synthetic data using loops"""
import os
from os import path as osp
import inspect

import pandas as pd
from scaper import generate_from_jams

from .logger import create_logger
from .post_process import _post_process_labels_file, get_labels_from_jams
from .soundscape import Soundscape
from .utils import create_folder, _check_random_state


class SoundscapesGenerator:
    """ Helps to generate set of Soundscapes
    Args:
        duration: float, the duration desired of the generated soundscapes
        fg_folder: str, path to the "foreground" folder. Contains one subfolder for each label
        bg_folder: str, path to the "background" folder. Contains one subfolder for each label
        ref_db: float, the dB reference of audio files (See scaper)
        samplerate: int, the sample rate desired of the generated soundscapes (be careful of the soundscape sample rate)
        random_state: np.random.RandomState or int, the random_state wanted to be able to reproduce the dataset
        delete_if_exists: bool, whether to delete existing files and folders created with the same name.
    """

    def __init__(
        self,
        duration,
        fg_folder,
        bg_folder,
        ref_db=-55,
        samplerate=16000,
        random_state=None,
        delete_if_exists=True,
        logger=None,
    ):
        self.duration = duration
        self.ref_db = ref_db
        self.fg_folder = fg_folder
        self.bg_folder = bg_folder
        self.samplerate = samplerate
        self.random_state = _check_random_state(random_state)
        self.delete_if_exists = delete_if_exists
        self.logger = logger
        if self.logger is None:
            self.logger = create_logger(
                __name__ + "/" + inspect.currentframe().f_code.co_name
            )

    def generate(
        self,
        number,
        out_folder,
        min_events=1,
        max_events=5,
        labels=("choose", []),
        source_files=("choose", []),
        sources_time=("const", 0),
        events_start=("truncnorm", 5.0, 2.0, 0.0, 10.0),
        events_duration=("uniform", 0.25, 10.0),
        snrs=("const", 30),
        pitch_shifts=("uniform", -3.0, 3.0),
        time_stretches=None,
        save_isolated_events=False,
        bg_labels=None,
        txt_file=True,
        start_from=0,
        **kwargs,
    ):
        """ Generate

        Args:
            number: int, number of audio clips to create.
            out_folder: str, path to extract generate file
            min_events: int, the minimum number of foreground events to add (pick at random uniformly).
            max_events: int, the maximum number of foreground events to add (pick at random uniformly).

            labels: tuple or list, distribution to choose foreground events or list of events.*
            source_files: tuple or list, distribution to choose source files or list of source files.*
            sources_time: tuple or list, distribution to choose source start time or list of sources start time.*
            events_start: tuple or list, distribution to choose events start time or list of events start time.*
            events_duration: tuple or list, distribution to choose events duation or list of events duration.*
            snrs: tuple or list, distribution to choose foreground to background SNRs or list of SNRs.*
            pitch_shifts: tuple or list, distribution to choose pitch shift or list of pitch shifts.*
            time_stretches: tuple or list, distribution to choose time stretches or list of time stretches.*
            save_isolated_events: bool, whether or not to save isolated events in a separate folder
            bg_labels: list, if None choose in all available files. If a name is given it has to match the name
                of a folder in 'background'. example: ["sins"]

            txt_file: bool, whether or not to save the .txt file.
            start_from: int, the number to start from if file already created.
            kwargs: arguments accepted by Scaper.generate

            * tuple is in the form of a distribution accepted by scaper.
        Returns:
            None
        """
        create_folder(out_folder)
        params = {
            "labels": labels,
            "source_files": source_files,
            "sources_time": sources_time,
            "events_start": events_start,
            "events_duration": events_duration,
            "snrs": snrs,
            "pitch_shifts": pitch_shifts,
            "time_stretches": time_stretches,
        }

        for cnt in range(number):
            self.logger.debug(
                "Generating soundscape: {:d}/{:d}".format(cnt + 1, number)
            )
            # create a scaper
            n_events = self.random_state.randint(min_events, max_events + 1)

            if start_from + cnt < 10:
                filename = "0" + str(start_from + cnt)
            else:
                filename = str(start_from + cnt)

            sc = Soundscape(
                self.duration,
                self.fg_folder,
                self.bg_folder,
                self.ref_db,
                self.samplerate,
                random_state=self.random_state,
                delete_if_exists=self.delete_if_exists,
            )
            sc.generate_one_bg_multi_fg(
                out_folder=out_folder,
                filename=filename,
                n_fg_events=n_events,
                **params,
                txt_file=txt_file,
                save_isolated_events=save_isolated_events,
                bg_labels=bg_labels,
                **kwargs,
            )
            if cnt % 200 == 0:
                self.logger.info(
                    f"generating {cnt} / {number} files (updated every 200)"
                )

    def generate_balance(
        self,
        number,
        out_folder,
        min_events,
        max_events,
        list_labels=None,
        save_isolated_events=False,
        start_from=0,
        snr=("uniform", 6, 30),
        pitch_shift=None,
        time_stretch=None,
        bg_labels=None,
        **kwargs,
    ):
        """ Generate landscapes by taking into account the probabilities of labels and their co-occurence
        Args:

            number: int, the number of files to generate
            out_folder: str, the path of the folder where to save the generated files
            min_events: int, the minimum number of events per files
            max_events: int, the maximum number of labels per file
            list_labels: list of available labels (in foreground)
            save_isolated_events: bool, whether or not to save isolated events in a subfolder
                (called <filename>_events by default)
            start_from: int, if already created file, will start the filenames at the specified number
            snr: tuple, tuple accepted by Scaper().add_event()
            pitch_shift: tuple, tuple accepted by Scaper().add_event()
            time_stretch: tuple, tuple accepted by Scaper().add_event()
            bg_labels: list, if None choose in all available files. If a name is given it has to match the name
                of a folder in 'background'. example: ["sins"]
            kwargs: parametes accepted by Scaper().generate()
        """
        create_folder(out_folder)
        cnt = 0
        if list_labels is None:
            list_labels = []
            for pth in os.listdir(self.fg_folder):
                if osp.isdir(osp.join(self.fg_folder, pth)):
                    if "non" in pth.lower() or "noff" in pth.lower():
                        for pattern in ["_nOn", "_nOff", "_non", "_noff"]:
                            pth = pth.replace(pattern, "")
                    list_labels.append(pth)
            list_labels = sorted(set(list_labels))
            self.logger.debug(f"list of labels: {list_labels}")

        for label in list_labels:
            self.logger.debug(
                "Generating soundscape: {:d}/{:d}".format(cnt + 1, number)
            )
            number_per_class = max(1, round(number // len(list_labels)))
            for i in range(number_per_class):
                sc = Soundscape(
                    self.duration,
                    self.fg_folder,
                    self.bg_folder,
                    self.ref_db,
                    self.samplerate,
                    random_state=self.random_state,
                    delete_if_exists=self.delete_if_exists,
                )
                if start_from + cnt < 10:
                    filename = "0" + str(start_from + cnt)
                else:
                    filename = str(start_from + cnt)
                if min_events == max_events:
                    n_events = min_events
                else:
                    n_events = self.random_state.randint(min_events, max_events)
                sc.generate_using_non_noff(
                    label=label,
                    list_labels=list_labels,
                    out_folder=out_folder,
                    filename=filename,
                    n_events=n_events,
                    save_isolated_events=save_isolated_events,
                    snr=snr,
                    pitch_shift=pitch_shift,
                    time_stretch=time_stretch,
                    bg_labels=bg_labels,
                    **kwargs,
                )
                if cnt % 200 == 0:
                    self.logger.info(
                        f"generating {cnt} / {number} files (updated every 200)"
                    )
                cnt += 1
        if cnt != number:
            self.logger.warn(
                f"The number of generated examples ({cnt}) is different from the number asked ({number}) "
                f"because of probabilities of events."
            )

    def generate_by_label_occurence(
        self,
        label_occurences,
        number,
        out_folder,
        min_events=0,
        max_events=None,
        save_isolated_events=False,
        start_from=0,
        snr=("uniform", 6, 30),
        pitch_shift=None,
        time_stretch=None,
        bg_labels=None,
        **kwargs,
    ):
        """ Generate landscapes by taking into account the probabilities of labels and their co-occurence
        Args:
            label_occurences: dict, parameters of labels occurences (foreground labels)
            number: int, the number of files to generate
            out_folder: str, the path of the folder where to save the generated files
            min_events: int, optional, the minimum number of events per files (default=0)
                (Be careful, if max_events in label_occurences params is less than this it will raise an error)
                If defined in the label_occurences dict, this parameter overwrites it.
            max_events: int, optional, if defined, overwrite the value in label_occurences if defined.
            save_isolated_events: bool, whether or not to save isolated events in a subfolder
                (called <filename>_events by default)
            start_from: int, if already created file, will start the filenames at the specified number
            snr: tuple, tuple accepted by Scaper().add_event()
            pitch_shift: tuple, tuple accepted by Scaper().add_event()
            time_stretch: tuple, tuple accepted by Scaper().add_event()
            bg_labels: list, if None choose in all available files. If a name is given it has to match the name
                of a folder in 'background'. example: ["sins"]
            kwargs: parametes accepted by Scaper().generate()
        Returns:

        Examples:
            An example of a JSON looks like this:
            event_occurences.json
            {
              "Alarm_bell_ringing": {
                "proba" : 0.6,
                "co-occurences": {
                  "max_events": 13,
                  "classes": [
                    "Alarm_bell_ringing",
                    "Dog",
                  ],
                  "probas": [
                    70,
                    30
                  ]
                },
              "Dog": {
                "prob" : 0.4,
                "co-occurences": {
                "max_events": 14,
                "classes": [
                  "Alarm_bell_ringing",
                  "Dog",
                ],
                "event_prob": [
                  25,
                  75,
                ]
              }
            }
        """
        create_folder(out_folder)
        cnt = 0
        for label in label_occurences.keys():
            self.logger.debug(
                "Generating soundscape: {:d}/{:d}".format(cnt + 1, number)
            )
            label_params = label_occurences[label]
            for i in range(round(number * label_params["proba"])):
                sc = Soundscape(
                    self.duration,
                    self.fg_folder,
                    self.bg_folder,
                    self.ref_db,
                    self.samplerate,
                    random_state=self.random_state,
                    delete_if_exists=self.delete_if_exists,
                )
                if start_from + cnt < 10:
                    filename = "0" + str(start_from + cnt)
                else:
                    filename = str(start_from + cnt)

                sc.generate_co_occurence(
                    co_occur_params=label_params["co-occurences"],
                    label=label,
                    out_folder=out_folder,
                    filename=filename,
                    min_events=min_events,
                    max_events=max_events,
                    save_isolated_events=save_isolated_events,
                    snr=snr,
                    pitch_shift=pitch_shift,
                    time_stretch=time_stretch,
                    bg_labels=bg_labels,
                    **kwargs,
                )
                if cnt % 200 == 0:
                    self.logger.info(
                        f"generating {cnt} / {number} files (updated every 200)"
                    )
                cnt += 1
        if cnt != number:
            self.logger.warn(
                f"The number of generated examples ({cnt}) is different from the number asked ({number}) "
                f"because of probabilities of events."
            )


def generate_df_from_jams(list_jams, post_process=True, background_label=False):
    if len(list_jams) == 0:
        raise IndexError(
            "Cannot generate df from JAMS, the list of jams given is empty"
        )

    final_df = pd.DataFrame()
    for jam_file in list_jams:
        fbase = osp.basename(jam_file)
        df, length = get_labels_from_jams(
            jam_file, background_label=background_label, return_length=True
        )

        if post_process:
            df, _ = _post_process_labels_file(df, length)

        df["filename"] = f"{osp.splitext(fbase)[0]}.wav"
        final_df = final_df.append(
            df[["filename", "onset", "offset", "event_label"]], ignore_index=True
        )

    final_df = final_df.sort_values(by=["filename", "onset"])
    return final_df


def generate_tsv_from_jams(
    list_jams, tsv_out, post_process=True, background_label=False
):
    """ In scaper.generate they create a txt file for each audio file.
    Using the same idea, we create a single tsv file with all the audio files and their labels.
    Args:
        list_jams: list, list of paths of JAMS files. Assume WAV files have the same name as JAMS files.
        tsv_out: str, path of the tsv to be saved
        post_process: bool, post_process removes small blanks, clean the overlapping same events in the labels and
        make the smallest event 250ms long.
        background_label: bool, include the background label in the annotations.
        # source_sep_path: str, the path to save the csv of separated source files. Assume

    Returns:
        None
    """
    create_folder(osp.dirname(tsv_out))
    final_df = generate_df_from_jams(list_jams, post_process, background_label)
    final_df.to_csv(tsv_out, sep="\t", index=False, float_format="%.3f")


def generate_files_from_jams(
    list_jams,
    out_folder,
    fg_path=None,
    bg_path=None,
    out_folder_jams=None,
    save_isolated_events=False,
    overwrite_exist_audio=False,
    **kwargs,
):
    """ Generate audio files from jams files generated by Scaper

    Args:
        list_jams: list, list of jams filepath generated by Scaper.
        out_folder: str, output path to save audio files.
        fg_path: str, the path to the foreground events, if not specified, should match what specified in the JAMS.
        bg_path: str, the path to the background events, if not specified, should match what specified in the JAMS.
        out_folder_jams: str, path to write the jams (could be modified by fg_path and bg_path for example),
            if None, jams not saved
        save_isolated_events: bool, whether or not to save isolated events in a separate folder
        overwrite_exist_audio: bool, whether to regenerate existing audio files or not
        kwargs: dict, scaper.generate_from_jams params (fg_path, bg_path, ...)
    Returns: None

    """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    logger.info(f"generating audio files to {out_folder}")
    create_folder(out_folder)
    for n, jam_file in enumerate(list_jams):
        logger.debug(jam_file)
        audiofile = osp.join(
            out_folder, f"{osp.splitext(osp.basename(jam_file))[0]}.wav"
        )
        if not osp.exists(audiofile) or overwrite_exist_audio:
            if out_folder_jams is not None:
                jams_outfile = osp.join(out_folder_jams, osp.basename(jam_file))
            else:
                jams_outfile = None
            generate_from_jams(
                jam_file,
                audiofile,
                fg_path=fg_path,
                bg_path=bg_path,
                jams_outfile=jams_outfile,
                save_isolated_events=save_isolated_events,
                **kwargs,
            )

        if n % 200 == 0:
            logger.info(f"generating {n} / {len(list_jams)} files (updated every 200)")
    logger.info("Done")



"""Post processing of desed synthetic generation of soundscapes"""
import glob
import inspect
import os
import shutil
from os import path as osp

import jams
import numpy as np
import pandas as pd
import soundfile as sf
from .logger import create_logger
from .utils import create_folder


def save_tsv(df, filepath):
    create_folder(osp.dirname(filepath))
    df = df.sort_values("filename")
    df.to_csv(filepath, index=False, sep="\t", float_format="%.3f")


def rm_high_polyphony(
    folder, max_polyphony=3, save_tsv_associated=None, pattern_sources="_events"
):
    """ Remove the files having a too high polyphony in the deignated folder

    Args:
        folder: str, path to the folder containing scaper generated sounds (JAMS files) in which to remove the files.
        max_polyphony: int, the maximum number of sounds that can be heard at the same time (polyphony).
        save_tsv_associated: str, optional, the path to generate the tsv files of associated sounds.
        pattern_sources: str, optional, the pattern that is added to the source to get isolated events
            (to be able to delete them if needed)
    Returns:
        None

    """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    # Select training
    i = 0
    df = pd.DataFrame(columns=["scaper", "bg", "fg"])
    fnames_to_rmv = []
    for jam_file in sorted(glob.glob(osp.join(folder, "*.jams"))):
        param = jams.load(jam_file)
        ann = param.annotations.search(namespace="scaper")[0]
        if ann["sandbox"]["scaper"]["polyphony_max"] <= max_polyphony:
            fg = [osp.basename(line.value["source_file"]) for line in ann.data]
            bg = osp.basename(ann.data[0].value["source_file"])
            fname = osp.basename(jam_file)
            df_tmp = pd.DataFrame(
                np.array([[fname, bg, ",".join(fg)]]), columns=["scaper", "bg", "fg"]
            )
            df = df.append(df_tmp, ignore_index=True)
            i += 1
        else:
            fnames_to_rmv.append(jam_file)
    if save_tsv_associated is not None:
        df.to_csv(save_tsv_associated, sep="\t", index=False)

    logger.warning(
        f"{i} files with less than {max_polyphony} overlapping events. Deleting others..."
    )
    for fname in fnames_to_rmv:
        names = glob.glob(osp.splitext(fname)[0] + ".*")
        for file in names:
            os.remove(file)
        dirs_sources = glob.glob(osp.splitext(fname)[0] + pattern_sources)
        for dir_path in dirs_sources:
            shutil.rmtree(dir_path)


def sanity_check(df, length_sec=None):
    """ Check that onset and offset are in the boundaries
    Args:
        df: pandas.DataFrame, dataframe defining 'onset' and 'offset' columns.
        length_sec: float, optional, if defined it is the maximum length of a file.

    Returns:
        pandas.DataFrame, the updated dataframe.
    """
    if not df["onset"].dtype == np.float and df["offset"].dtype == np.float:
        df[["onset", "offset"]] = df[["onset", "offset"]].astype(float)
    if length_sec is not None:
        df["offset"].clip(upper=length_sec, inplace=True)
    df["onset"].clip(lower=0, inplace=True)
    df = df.round(3)
    return df


def get_data(file, wav_file=None, background_label=False):
    """ Get annotation of a file (txt or JAMS) and check the correspondance with a wav file (created by Scaper).
    Args:
        file: str, path of the .txt or .jams file.
        wav_file: str, path of the wav file associated with the 'file'.
        background_label: bool, whether to get the background as a label or not.

    Returns:

    """
    if wav_file is not None:
        data, sr = sf.read(wav_file)
        length_sec = data.shape[0] / sr
    else:
        length_sec = None

    fn, ext = osp.splitext(file)
    if ext == ".txt":
        if background_label:
            raise NotImplementedError(
                "Impossible to add the background event from the txt file. "
                "Information not in the txt file"
            )
        df = pd.read_csv(file, sep="\t", names=["onset", "offset", "event_label"])
    elif ext == ".jams":
        df = get_labels_from_jams(file, background_label)
    else:
        raise NotImplementedError(
            "Only txt and jams generated by Scaper can be loaded with get_data"
        )

    return df, length_sec


def _post_process_labels_file(
    df_ann, length_sec=None, min_dur_event=0.250, min_dur_inter=0.150, rm_non_noff=True
):
    """ Check the annotations,
        * Merge overlapping annotations of the same class
        * Merge overlapping annotations having less than 150ms between them (or 400ms between the onsets).
        * Make minimum length of events = 250ms.
    Args:
        df_ann: pd.DataFrame object, containing the annotations to post_process
        length_sec: float, duration of the file to post_process
        min_dur_event: float, optional in sec, minimum duration of an event
        min_dur_inter: float, optional in sec, minimum duration between 2 events
        rm_non_noff: bool, whether to delete the additional _nOn _nOff at the end of labels.

    Returns:

    """

    df = df_ann.copy()
    if rm_non_noff:
        df["event_label"] = df["event_label"].apply(
            lambda x: str(x).replace("_nOff", "").replace("_nOn", "")
        )
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    fix_count = 0
    df = sanity_check(df, length_sec)
    df = df.sort_values("onset")
    for class_name in df["event_label"].unique():
        logger.debug(class_name)
        i = 0
        while i is not None:
            indexes = df[df["event_label"] == class_name].index
            ref_onset = df.loc[indexes[i], "onset"]
            ref_offset = df.loc[indexes[i], "offset"]
            if ref_offset - ref_onset < min_dur_event:
                ref_offset = ref_onset + min_dur_event
                # Too short event, and at the offset (onset sorted),
                # so if it overlaps with others, they are also too short.
                if ref_offset > length_sec:
                    df = df.drop(indexes[i:])
                    fix_count += len(indexes[i:])
                    break
                else:
                    df.loc[indexes[i], "offset"] = ref_onset + min_dur_event
            j = i + 1
            while j < len(indexes):
                if df.loc[indexes[j], "offset"] < ref_offset:
                    df = df.drop(indexes[j])
                    logger.debug("Merging overlapping annotations")
                    fix_count += 1
                elif df.loc[indexes[j], "onset"] - ref_offset < min_dur_inter:
                    df.loc[indexes[i], "offset"] = df.loc[indexes[j], "offset"]
                    ref_offset = df.loc[indexes[j], "offset"]
                    df = df.drop(indexes[j])
                    logger.debug("Merging consecutive annotation with pause" + "<150ms")
                    fix_count += 1
                elif (
                    df.loc[indexes[j], "onset"] - ref_onset
                    < min_dur_event + min_dur_inter
                ):
                    df.loc[indexes[i], "offset"] = df.loc[indexes[j], "offset"]
                    ref_offset = df.loc[indexes[j], "offset"]
                    df = df.drop(indexes[j])
                    logger.debug(
                        "Merging consecutive annotations" + " with onset diff<400ms"
                    )
                    fix_count += 1
                else:
                    # Quitting the loop
                    break
                j += 1
            i += 1
            if i >= len(df[df["event_label"] == class_name].index):
                i = None
    df = df.sort_values("onset")
    return df, fix_count


def post_process_df_labels(
    df,
    files_duration=None,
    output_tsv=None,
    min_dur_event=0.250,
    min_dur_inter=0.150,
    rm_nOn_nOff=True,
):
    """ clean the .txt files of each file. It is the same processing as the real data
        - overlapping events of the same class are mixed
        - if silence < 150ms between two conscutive events of the same class, they are mixed
        - if event < 250ms, the event lasts 250ms

        Args:
            df: pd.DataFrame, dataframe of annotations containing columns ["filename", "onset", "offset", "event_label"]
            files_duration: pd.DataFrame or float, dataframe containing columns ["filename", "duration"]
                indicating the lengh of a file.
                or float being the length of all the files if all the files have the same duration.
            output_tsv: str, optional, tsv with all the annotations concatenated
            min_dur_event: float, optional in sec, minimum duration of an event
            min_dur_inter: float, optional in sec, minimum duration between 2 events
            rm_nOn_nOff: bool, whether to delete the additional _nOn _nOff at the end of labels.

        Returns:
            None
        """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    fix_count = 0
    logger.info(
        "Correcting annotations ... \n"
        "* annotations with negative duration will be removed\n"
        + "* annotations with duration <250ms will be extended on the offset side)"
    )

    result_df = pd.DataFrame()
    for fn in df.filename.unique():
        logger.debug(fn)
        if files_duration is not None:
            if type(files_duration) is pd.DataFrame:
                length_sec = files_duration[files_duration.filename == fn].duration
            elif type(files_duration) in [float, int]:
                length_sec = files_duration
            else:
                raise TypeError("files duration is pd.DataFrame or a float only")
        else:
            length_sec = None
        df_ann, fc = _post_process_labels_file(
            df[df.filename == fn], length_sec, min_dur_event, min_dur_inter, rm_nOn_nOff
        )
        fix_count += fc

        result_df = result_df.append(
            df_ann[["filename", "onset", "offset", "event_label"]], ignore_index=True
        )

    if output_tsv:
        save_tsv(result_df, output_tsv)

    logger.info(f"================\nFixed {fix_count} problems\n================")
    return result_df


def post_process_txt_labels(
    txtdir,
    wavdir=None,
    output_folder=None,
    output_tsv=None,
    min_dur_event=0.250,
    min_dur_inter=0.150,
    background_label=False,
    rm_nOn_nOff=True,
):
    """ clean the .txt files of each file. It is the same processing as the real data
    - overlapping events of the same class are mixed
    - if silence < 150ms between two conscutive events of the same class, they are mixed
    - if event < 250ms, the event lasts 250ms

    Args:
        txtdir: str, directory path where the XXX.txt files are.
        wavdir: str, directory path where the associated XXX.wav audio files are (associated with .txt files)
        output_folder: str, optional, folder in which to put the checked files
        output_tsv: str, optional, tsv with all the annotations concatenated
        min_dur_event: float, optional in sec, minimum duration of an event
        min_dur_inter: float, optional in sec, minimum duration between 2 events
        background_label: bool, whether to include the background label in the annotations.
        rm_nOn_nOff: bool, whether to delete the additional _nOn _nOff at the end of labels.

    Returns:
        None
    """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    if wavdir is None:
        wavdir = txtdir
    fix_count = 0
    logger.info(
        "Correcting annotations ... \n"
        "* annotations with negative duration will be removed\n"
        + "* annotations with duration <250ms will be extended on the offset side)"
    )

    if output_folder is not None:
        create_folder(output_folder)

    df_single = pd.DataFrame()  # only useful if output_csv defined

    if background_label:
        list_files = glob.glob(osp.join(txtdir, "*.jams"))
    else:
        list_files = glob.glob(osp.join(txtdir, "*.txt"))
        if len(list_files) == 0:
            list_files = glob.glob(osp.join(txtdir, "*.jams"))

    out_extension = ".txt"
    for fn in list_files:
        logger.debug(fn)
        df, length_sec = get_data(
            fn,
            osp.join(wavdir, osp.splitext(osp.basename(fn))[0] + ".wav"),
            background_label=background_label,
        )

        df, fc = _post_process_labels_file(
            df, length_sec, min_dur_event, min_dur_inter, rm_nOn_nOff
        )
        fix_count += fc

        if output_folder is not None:
            filepath = os.path.splitext(os.path.basename(fn))[0] + out_extension
            df[["onset", "offset", "event_label"]].to_csv(
                osp.join(output_folder, filepath), header=False, index=False, sep="\t"
            )
        if output_tsv is not None:
            df["filename"] = osp.join(osp.splitext(osp.basename(fn))[0] + ".wav")
            df_single = df_single.append(
                df[["filename", "onset", "offset", "event_label"]], ignore_index=True
            )

    if output_tsv:
        save_tsv(df_single, output_tsv)

    logger.info(f"{fix_count} problems Fixed")


def get_labels_from_jams(jam_file, background_label=False, return_length=False):
    tsv_data = []
    param = jams.load(jam_file)
    ann = param["annotations"][0]
    for obs in ann.data:
        if obs.value["role"] == "foreground" or (
            background_label and obs.value["role"] == "background"
        ):
            tsv_data.append([obs.time, obs.time + obs.duration, obs.value["label"]])
    df = pd.DataFrame(tsv_data, columns=["onset", "offset", "event_label"])

    if return_length:
        return df, ann.duration
    else:
        return df



# -*- coding: utf-8 -*-
import functools
import inspect
import numbers

import jams
import numpy as np
import os
import os.path as osp
import shutil
import pprint
import requests
import sys

from .logger import create_logger, DesedError


def _check_random_state(seed):
    """Turn seed into a np.random.RandomState instance

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    elif isinstance(seed, (numbers.Integral, np.integer, int)):
        return np.random.RandomState(seed)
    elif isinstance(seed, np.random.RandomState):
        return seed
    else:
        raise ValueError(
            "%r cannot be used to seed a numpy.random.RandomState" " instance" % seed
        )


def create_folder(folder, exist_ok=True, delete_if_exists=False):
    """ Create folder (and parent folders) if not exists.

    Args:
        folder: str, path of folder(s) to create.
        delete_if_exists: bool, True if you want to delete the folder when exists

    Returns:
        None
    """
    if delete_if_exists:
        if os.path.exists(folder):
            shutil.rmtree(folder)
            os.mkdir(folder)

    os.makedirs(folder, exist_ok=exist_ok)


pp = pprint.PrettyPrinter()


def pprint(x):
    pp.pprint(x)


def choose_cooccurence_class(co_occur_params, random_state=None):
    """ Choose another class given a dictionary of parameters (from an already specified class).
    Args:
        co_occur_params: dict, define the parameters of co-occurence of classes
            Example of co_occur_params dictionnary::
                {
                  "max_events": 13,
                  "classes": [
                    "Alarm_bell_ringing",
                    "Dog",
                  ],
                  "probas": [
                    70,
                    30
                  ]
                }
            classes and probas maps each others
        random_state: int, or RandomState object
    Returns:
        str, the class name.
    """
    if random_state is not None:
        random_state = _check_random_state(random_state)
        chosen_class = random_state.choice(
            co_occur_params["classes"], p=co_occur_params["probas"]
        )
    else:
        chosen_class = np.random.choice(
            co_occur_params["classes"], p=co_occur_params["probas"]
        )
    return chosen_class


def change_snr(jams_path, db_change):
    """ Modify the background SNR of a JAMS generated by scaper
    Args:
        jams_path: str, jams path of file created by Scaper
        db_change: float, Signal to noise ratio (SNR), in dB to be added

    Returns:
        jam_obj the jams object that has been modified
    """
    jams_obj = jams.load(jams_path)
    ann = jams_obj.annotations.search(namespace="scaper")[0]
    for cnt, obs in enumerate(ann.data):
        if obs.value["role"] == "foreground":
            old_snr = ann.data[cnt].value["snr"]
            # Changing manually the jams to reduce the SNR (note that snr can be negative)
            ann.data[cnt].value["snr"] = old_snr + db_change

    return jams_obj


def modify_fg_onset(jams_path, slice_seconds):
    """ Add a value foreground onset of a JAMS generated by scaper (containing a single event)
    Args:
        jams_path: str, the name of the JAMS file to change the background SNR
        slice_seconds: float, value in seconds, value to be added to previous onset

    Returns:
        jams object that has been modified
    """
    jams_obj = jams.load(jams_path)
    ann = jams_obj.annotations.search(namespace="scaper")[0]
    data = ann.data
    for cnt, obs in enumerate(data):
        if obs.value["role"] == "foreground":
            onset = obs.value["event_time"]
            # Checking the new onset is possible
            if onset + slice_seconds > ann.duration:
                raise DesedError(
                    f"The new onset is not valid: {onset + slice_seconds} > {ann.duration}, "
                    f"for file: {jams_path}"
                )
            elif onset + slice_seconds < 0:
                raise DesedError(
                    f"The new onset is not valid: {onset + slice_seconds} < 0, "
                    f"for file: {jams_path}"
                )
            else:
                # Change source time by adding the added value specified
                ann.data[cnt].value["event_time"] = onset + slice_seconds
                # Todo, this is tricky, because it is an object, find a better way to do that with Scaper
                new_obs = ann.data[cnt]._replace(time=onset + slice_seconds)
                del ann.data[cnt]
                ann.data.add(new_obs)

    return jams_obj


def modify_jams(list_jams, modify_function, out_dir=None, **kwargs):
    """ Function to modify jams files
    Args:
        modify_function: function, a function that takes (jams_path, **kwargs) as input, and
            return the modified jams object (jams.JAMS)
        list_jams: list, the list of jams_path to be modified
        out_dir: str, the path of the directory of the new jams. If not defined, overwrite the jam_file

        **kwargs: arguments that'll be given to modify_function

    Returns: None

    Examples:
        There are two ways of using this function, (example with snr diminution):
        >>> jams_to_modify = ["material/5.jams"]
        1-
        >>> modify_jams(jams_to_modify, change_snr, db_added=-6)
        2- Recommended
        >>> decrease_snr = functools.partial(change_snr, db_added =-6)
        >>> modify_jams(jams_to_modify, decrease_snr)
    """
    logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
    create_folder(out_dir)
    new_list_jams = []
    for jam_file in list_jams:
        logger.debug(jam_file)
        jam_obj = modify_function(jam_file, **kwargs)

        if out_dir is not None:
            out_jams = osp.join(out_dir, os.path.basename(jam_file))
        else:
            out_jams = jam_file
        jam_obj.save(out_jams)
        new_list_jams.append(out_jams)

    return new_list_jams


def download_file_from_url(url, target_destination):
    """ Download a file from a URL.

    Args:
        url: str, URL to be download.
        target_destination: str, the file in which to output the content of the URL.

    Returns:

    """
    print(f"Downloading {os.path.basename(url)}")
    response = requests.get(url, stream=True)
    handle = open(target_destination, "wb")
    total_length = response.headers.get("content-length")
    if total_length is None:  # no content length header
        handle.write(response.content)
    else:
        size_dl = 0
        total_length = int(total_length)
        for chunk in response.iter_content(chunk_size=512):
            if chunk:  # filter out keep-alive new chunks
                size_dl += len(chunk)
                handle.write(chunk)
                done = int(50 * size_dl / total_length)
                sys.stdout.write(f"\r[{'=' * done}{' ' * (50-done)}]")
                sys.stdout.flush()
    print("\n")



"""Sounscape is a subclass of scaper.Scaper to fulfill our specific needs and default values"""
import glob
import inspect
import numpy as np
import os
import shutil
import warnings
from os import path as osp

import scaper
import soundfile as sf

from .logger import create_logger, DesedWarning, DesedError
from .utils import choose_cooccurence_class, create_folder


class Soundscape(scaper.Scaper):
    """ Initialize scaper object with a reference dB and sr.
            Args:
                duration: float, in seconds, the duration of the generated audio clip.
                fg_path: str, path of foreground files (be careful, need subfolders, one per class or group)
                bg_path: str, path of background files (be careful, need subfolders, one per group)
                ref_db: float, the reference dB of the clip.
                samplerate: int, the sr of the final soundscape
                delete_if_exists: bool, whether to delete existing files and folders created with the same name.

            Returns:
                scaper.Scaper object
    """

    def __init__(
        self,
        duration,
        fg_path,
        bg_path,
        ref_db=-55,
        samplerate=16000,
        random_state=None,
        delete_if_exists=True,
    ):
        super(Soundscape, self).__init__(
            duration, fg_path, bg_path, random_state=random_state
        )
        self.ref_db = ref_db
        self.sr = samplerate
        if self.ref_db is not None:
            self.ref_db = ref_db
        if self.sr is not None:
            self.sr = samplerate
        self.delete_if_exists = delete_if_exists

    def add_random_background(self, label=None):
        """ Add a random background to a scaper object
        Args:
            label: str or list, possible labels are names the subfolders of self.bg_path. None can use them all.
        """
        # If str or None, keep it like this
        if label is not None:
            if isinstance(label, list):
                bg_label = self.random_state.choice(label)
            elif isinstance(label, str):
                bg_label = label
            else:
                raise NotImplementedError(
                    "Background label can only be a list of available labels or a string"
                )
        else:
            bg_label = "*"
        chosen_file = self._choose_file(osp.join(self.bg_path, bg_label))
        file_duration = sf.info(chosen_file).duration
        starting_source = min(
            self.random_state.rand() * file_duration,
            max(file_duration - self.duration, 0),
        )
        self.add_background(
            label=("const", chosen_file.split("/")[-2]),
            source_file=("const", chosen_file),
            source_time=("const", starting_source),
        )

    def add_fg_event_non_noff(
        self,
        label,
        snr=("uniform", 6, 30),
        pitch_shift=None,
        time_stretch=None,
        event_duration_min=0.25,
    ):
        """ add a single event to a scaper object given a class. Take into account if the event has an onset or offset
        Args:
            label: str, label of the event to add.
            snr: tuple, tuple accepted by Scaper().add_event()
            pitch_shift: tuple, tuple accepted by Scaper().add_event()
            time_stretch: tuple, tuple accepted by Scaper().add_event()
            event_duration_min: float, minimum duration of a foreground event (in second)

        Returns:
            sc, scaper.Scaper object with the event added.

        """
        logger = create_logger(__name__ + "/" + inspect.currentframe().f_code.co_name)
        label_path = os.path.join(self.fg_path, label)
        assert osp.exists(
            label_path
        ), f"The label provided ({label}) does not point to a valid folder: {label_path}"
        chosen_file = self._choose_file(os.path.join(self.fg_path, label))
        file_duration = round(
            sf.info(chosen_file).duration, 6
        )  # because Scaper uses sox with truncate 6 digits
        if "_nOn_nOff" in label:
            # If no onset and offset, the file should be bigger than the duration of the file
            logger.debug("no onset/offset")
            if file_duration - self.duration <= 0:
                warnings.warn(
                    "Event without onset and offset added not for the full time of the audio soundscape"
                )
                source_start = 0
            else:
                source_start = self.random_state.uniform(
                    0, file_duration - self.duration
                )

            self.add_event(
                label=("const", label),
                source_file=("const", chosen_file),
                source_time=("const", source_start),
                event_time=("const", 0),
                event_duration=("const", self.duration),
                snr=snr,
                pitch_shift=pitch_shift,
                time_stretch=time_stretch,
            )
        elif "_nOn" in label:
            logger.debug("no onset")
            if file_duration - event_duration_min <= 0:
                source_start = 0
            else:
                source_start = self.random_state.uniform(
                    0, file_duration - event_duration_min
                )
            self.add_event(
                label=("const", label),
                source_file=("const", chosen_file),
                source_time=("const", source_start),
                event_time=("const", 0),
                event_duration=(
                    "const",
                    np.minimum(self.duration, file_duration - source_start),
                ),
                snr=snr,
                pitch_shift=pitch_shift,
                time_stretch=time_stretch,
            )
        elif "_nOff" in label:
            logger.debug("no offset")
            event_start = self.random_state.uniform(
                max(0, self.duration - file_duration),
                self.duration - event_duration_min,
            )
            event_length = self.duration - event_start
            self.add_event(
                label=("const", label),
                source_file=("const", chosen_file),
                source_time=("const", 0),
                event_time=("const", event_start),
                event_duration=("const", event_length),
                snr=snr,
                pitch_shift=pitch_shift,
                time_stretch=time_stretch,
            )
        else:
            logger.debug("onset offset")
            if file_duration > self.duration:
                if file_duration // self.duration > 2:
                    choice = np.random.choice(
                        ["onset", "middle", "offset"], p=[0.375, 0.25, 0.375]
                    )
                else:
                    choice = np.random.choice(["onset", "offset"])
                logger.debug(f"choice: {choice}")
                if choice == "onset":
                    event_start = self.random_state.uniform(
                        0, self.duration - event_duration_min
                    )
                    length_possible = self.duration - event_start
                    event_length = (
                        file_duration
                        if file_duration < length_possible
                        else length_possible
                    )
                    source_start = 0
                elif choice == "offset":
                    event_start = 0
                    event_length = self.random_state.uniform(
                        0, self.duration - event_duration_min
                    )
                    source_start = file_duration - event_length
                else:
                    source_start = self.random_state.uniform(
                        self.duration, file_duration - self.duration
                    )
                    event_start = 0
                    event_length = self.duration
            else:
                event_start = self.random_state.uniform(
                    0, self.duration - event_duration_min
                )
                source_start = 0
                length_possible = self.duration - event_start
                event_length = (
                    file_duration
                    if file_duration < length_possible
                    else length_possible
                )
            logger.debug(
                f"event_start: {event_start}, length: {event_length}, source_start: {source_start}, "
                f"file_duration: {file_duration}"
            )

            self.add_event(
                label=("const", label),
                source_file=("const", chosen_file),
                source_time=("const", source_start),
                event_time=("const", event_start),
                event_duration=("const", event_length),
                snr=snr,
                pitch_shift=pitch_shift,
                time_stretch=time_stretch,
            )
        return self

    def _choose_file(self, class_path, non_noff=False):
        """ Choose randomly a file of a given class.
        Args:
            class_path: str, path of the class containing all the files of a certain class.

        Returns:
            str, path of the file.
        """
        event_files = sorted(glob.glob(os.path.join(class_path, "*")))
        if non_noff:
            event_files.append(glob.glob(os.path.join(class_path + "_nOn", "*")))
            event_files.append(glob.glob(os.path.join(class_path + "_nOff", "*")))
            event_files.append(glob.glob(os.path.join(class_path + "_nOn_nOff", "*")))
            event_files = sorted(event_files)

        event_files = [f for f in event_files if os.path.isfile(f)]
        assert len(event_files) > 0, (
            f"no event files to be chosen in this path: {os.path.join(class_path, '*')}"
            f" (pattern used by glob)"
        )
        ind = self.random_state.randint(0, len(event_files))
        return event_files[ind]

    def _remove(self, path):
        if osp.exists(path):
            if osp.isdir(path):
                shutil.rmtree(path)
            elif osp.isfile(path):
                os.remove(path)
            else:
                raise NotImplementedError("Can only remove files or folders")

    def generate_co_occurence(
        self,
        co_occur_params,
        label,
        out_folder,
        filename,
        min_events=1,
        max_events=None,
        reverb=None,
        save_isolated_events=False,
        snr=("uniform", 6, 30),
        pitch_shift=None,
        time_stretch=None,
        bg_labels=None,
        **kwargs,
    ):
        """ Generate a single file, using the information of onset or offset present
        (see DESED dataset and folders in soundbank foreground)
        Args:
            co_occur_params: dict, dict containing information about how to mix classes,
                and the probability of each class.
            label: str, the main foreground label of the generated file.
            out_folder: str, path to extract generate file
            filename: str, name of the generated file, without extension (.wav, .jams and .txt will be created)
            min_events: int, optional, the minimum number of events per files (default=1, >= 1)
                (Be careful, if max_events in label_occurences params is less than this it will raise an error)
                If defined in the label_occurences dict, this parameter corresponds to the number of cooccurences.
            max_events: int, optional, if defined, overwrite the value in label_occurences if defined
                (in label_occurences this parameter corresponds to the number of cooccurences).
            reverb: float, the reverb to be applied to the foreground events
            save_isolated_events: bool, whether or not to save isolated events in a subfolder
                (called <filename>_events by default)
            snr: tuple, tuple accepted by Scaper().add_event()
            pitch_shift: tuple, tuple accepted by Scaper().add_event()
            time_stretch: tuple, tuple accepted by Scaper().add_event()
            bg_labels: list or str, if None choose in all available files.
                If a name or list is given it has to match the name of a folder in 'background'. example: "sins"
            kwargs: arguments accepted by Scaper.generate
        Returns:
            None

        Examples:
            Example of co_occur_params dictionary::
                {
                  "max_events": 13,
                  "classes": [
                    "Alarm_bell_ringing",
                    "Dog",
                  ],
                  "probas": [
                    70,
                    30
                  ]
                }
            prob is the probability of this class (not used here)
        """
        create_folder(out_folder)
        self.add_random_background(bg_labels)

        # add main event, non_noff stands for no onset and no offset (accept label to have _nOn or _nOff specified).
        self.add_fg_event_non_noff(
            label, snr=snr, pitch_shift=pitch_shift, time_stretch=time_stretch
        )

        if max_events is None:
            max_events = co_occur_params.get("max_events")
            if max_events is None:
                raise DesedError("max_events has to be specified")
        else:
            max_events = max_events - 1

        if min_events is None:
            min_events = co_occur_params.get("min_events")
            if min_events is None:
                raise DesedError(
                    "min_events has to be specified in generate co occurence or in params"
                )
        else:
            min_events = min_events - 1

        # add random number of foreground events
        if min_events == max_events:
            n_events = min_events
        else:
            n_events = self.random_state.randint(min_events, max_events)
        for _ in range(n_events):
            chosen_class = choose_cooccurence_class(
                co_occur_params, random_state=self.random_state
            )
            self.add_fg_event_non_noff(
                chosen_class,
                snr=snr,
                pitch_shift=pitch_shift,
                time_stretch=time_stretch,
            )

        # Just in case an extension has been added
        ext = osp.splitext(filename)[-1]
        if ext in [".wav", ".jams", ".txt"]:
            filename = osp.splitext(filename)[0]

        # generate
        audio_file = osp.join(out_folder, f"{filename}.wav")
        jams_file = osp.join(out_folder, f"{filename}.jams")
        txt_file = osp.join(out_folder, f"{filename}.txt")

        if self.delete_if_exists:
            self._remove(audio_file)
            self._remove(jams_file)
            self._remove(txt_file)

        # To get isolated events in a subfolder
        isolated_events_path = kwargs.get("isolated_events_path")
        if save_isolated_events:
            if isolated_events_path is None:
                isolated_events_path = osp.join(out_folder, f"{filename}_events")
            if self.delete_if_exists:
                self._remove(isolated_events_path)
            else:
                if osp.exists(isolated_events_path):
                    warnings.warn(
                        f"The folder {isolated_events_path} already exists, it means there could be some "
                        f"unwanted audio files from previous generated audio files in it.",
                        DesedWarning,
                    )

        self.generate(
            audio_file,
            jams_file,
            reverb=reverb,
            txt_path=txt_file,
            save_isolated_events=save_isolated_events,
            isolated_events_path=isolated_events_path,
            **kwargs,
        )

    def generate_using_non_noff(
        self,
        label,
        list_labels,
        out_folder,
        filename,
        n_events,
        reverb=None,
        save_isolated_events=False,
        snr=("uniform", 6, 30),
        pitch_shift=None,
        time_stretch=None,
        bg_labels=None,
        **kwargs,
    ):
        """ Generate a single file, using the information of onset or offset present
        (see DESED dataset and folders in soundbank foreground)
        Args:
            label: str, the main foreground label of the generated file.
            list_labels: list, the list of available labels
            out_folder: str, path to extract generate file
            filename: str, name of the generated file, without extension (.wav, .jams and .txt will be created)
            n_events: int, the of events in the soundscape
            reverb: float, the reverb to be applied to the foreground events
            save_isolated_events: bool, whether or not to save isolated events in a subfolder
                (called <filename>_events by default)
            snr: tuple, tuple accepted by Scaper().add_event()
            pitch_shift: tuple, tuple accepted by Scaper().add_event()
            time_stretch: tuple, tuple accepted by Scaper().add_event()
            bg_labels: list, if None choose in all available files. If a name is given it has to match the name
                of a folder in 'background'. example: ["sins"]
            kwargs: arguments accepted by Scaper.generate
        Returns:
            None

        """
        create_folder(out_folder)
        self.add_random_background(bg_labels)

        if n_events > 0:
            # add main event, non_noff stands for no onset and no offset (accept label to have _nOn or _nOff specified).
            self.add_fg_event_non_noff(
                label, snr=snr, pitch_shift=pitch_shift, time_stretch=time_stretch
            )

            for _ in range(n_events - 1):
                chosen_class = self.random_state.choice(list_labels)
                self.add_fg_event_non_noff(
                    chosen_class,
                    snr=snr,
                    pitch_shift=pitch_shift,
                    time_stretch=time_stretch,
                )

        # Just in case an extension has been added
        ext = osp.splitext(filename)[-1]
        if ext in [".wav", ".jams", ".txt"]:
            filename = osp.splitext(filename)[0]

        # generate
        audio_file = osp.join(out_folder, f"{filename}.wav")
        jams_file = osp.join(out_folder, f"{filename}.jams")
        txt_file = osp.join(out_folder, f"{filename}.txt")

        if self.delete_if_exists:
            self._remove(audio_file)
            self._remove(jams_file)
            self._remove(txt_file)

        # To get isolated events in a subfolder
        isolated_events_path = kwargs.get("isolated_events_path")
        if save_isolated_events:
            if isolated_events_path is None:
                isolated_events_path = osp.join(out_folder, f"{filename}_events")
            if self.delete_if_exists:
                self._remove(isolated_events_path)
            else:
                if osp.exists(isolated_events_path):
                    warnings.warn(
                        f"The folder {isolated_events_path} already exists, it means there could be some "
                        f"unwanted audio files from previous generated audio files in it.",
                        DesedWarning,
                    )

        self.generate(
            audio_file,
            jams_file,
            reverb=reverb,
            txt_path=txt_file,
            save_isolated_events=save_isolated_events,
            isolated_events_path=isolated_events_path,
            **kwargs,
        )

    def generate_one_bg_multi_fg(
        self,
        out_folder,
        filename,
        n_fg_events,
        labels=("choose", []),
        source_files=("choose", []),
        sources_time=("const", 0),
        events_start=("truncnorm", 5.0, 2.0, 0.0, 10.0),
        events_duration=("uniform", 0.25, 10.0),
        snrs=("uniform", 6, 30),
        pitch_shifts=("uniform", -3.0, 3.0),
        time_stretches=("uniform", 1, 1),
        reverb=0.1,
        txt_file=True,
        save_isolated_events=False,
        isolated_events_path=None,
        bg_labels=None,
        **kwargs,
    ):
        """ Generate a clip with a background file and multiple foreground files.
        Args:
            out_folder: str, path to extract generate file
            filename: str, name of the generated file, without extension (.wav, .jams and .txt will be created)
            n_fg_events: int, the number of foreground events to add
            labels: tuple or list, distribution to choose foreground events or list of events.*
            source_files: tuple or list, distribution to choose source files or list of source files.*
            sources_time: tuple or list, distribution to choose source start time or list of sources start time.*
            events_start: tuple or list, distribution to choose events start time or list of events start time.*
            events_duration: tuple or list, distribution to choose events duation or list of events duration.*
            snrs: tuple or list, distribution to choose foreground to background SNRs or list of SNRs.*
            pitch_shifts: tuple or list, distribution to choose pitch shift or list of pitch shifts.*
            time_stretches: tuple or list, distribution to choose time stretches or list of time stretches.*
            reverb: float, the reverb to be applied to the foreground events
            txt_file: bool, whether or not to save the .txt file.
            save_isolated_events: bool, whether or not to save isolated events in a separate folder
            isolated_events_path: str, only useful when save_isolated_events=True. Give the path to the events folders.
                If None, a folder is created next to the audio files.
            bg_labels: list, if None choose in all available files. If a name is given it has to match the name
                of a folder in 'background'. example: ["sins"]
            kwargs: arguments accepted by Scaper.generate

            * All arguments with asterix, if tuple given, see Scaper for distribution allowed.
        Returns:
            None
        """
        create_folder(out_folder)
        self.add_random_background(bg_labels)

        params = {
            "label": labels,
            "source_file": source_files,
            "source_time": sources_time,
            "event_time": events_start,
            "event_duration": events_duration,
            "snr": snrs,
            "pitch_shift": pitch_shifts,
            "time_stretch": time_stretches,
        }
        # Make sure that if we give a list of tuple for a parameter that the length of the list
        # is matching the number of foreground events
        for i in range(n_fg_events):
            event_params = {}
            for key in params:
                if params[key] is None or isinstance(params[key], tuple):
                    param = params[key]
                elif type(params[key]) is list:
                    assert len(params[key]) == n_fg_events
                    if not isinstance(params[key][i], tuple):
                        param = ("const", params[key][i])
                    else:
                        param = params[key][i]
                else:
                    raise NotImplementedError(
                        "Params of events is tuple(same for all) or "
                        "list (different for each event)"
                    )
                event_params[key] = param

            self.add_event(**event_params)

        # generate
        audiofile = osp.join(out_folder, f"{filename}.wav")
        jamsfile = osp.join(out_folder, f"{filename}.jams")

        if self.delete_if_exists:
            self._remove(audiofile)
            self._remove(jamsfile)

        if txt_file:
            # Can be useless if you want background annotation as well, see post_processing_annotations.
            txtfile = osp.join(out_folder, f"{filename}.txt")
            if self.delete_if_exists:
                self._remove(txtfile)
        else:
            txtfile = None

        if save_isolated_events:
            if isolated_events_path is None:
                isolated_events_path = osp.join(out_folder, f"{filename}_events")
            if self.delete_if_exists:
                self._remove(isolated_events_path)

        self.generate(
            audio_path=audiofile,
            jams_path=jamsfile,
            reverb=reverb,
            txt_path=txtfile,
            save_isolated_events=save_isolated_events,
            isolated_events_path=isolated_events_path,
            **kwargs,
        )



import os
from setuptools import setup
from setuptools import find_packages


def package_file(fname):
    return os.path.relpath(os.path.join(os.path.dirname(__file__), fname))


with open("README.md", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name='desed',
    version='1.3.1',
    description="DESED dataset utils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Nicolas Turpault",
    author_email="turpaultn@gmail.com",
    url="https://github.com/turpaultn/DESED",
    license='MIT',
    package_dir={"": package_file("desed")},
    packages=find_packages(package_file("desed")),
    install_requires=[
        "scaper >= 1.3.5",
        "numpy >= 1.15.4",
        "pandas >= 0.24.0",
        "dcase-util >= 0.2.11",
        "youtube-dl >= 2021.2.22",  # Better with conda install and conda-forge channel
        "soundfile >= 0.10.1",
        "jams >= 0.3.4",
        "tqdm >= 4.29.1",
        "requests >= 2.21.0"
    ]
)



from .download import (
    download_real,
    download_desed_soundbank,
    download_fuss,
    download_fsd50k,
)
from .generate_synthetic import (
    SoundscapesGenerator,
    generate_files_from_jams,
    generate_tsv_from_jams,
    generate_df_from_jams,
)
from . import post_process, utils
